/*
 see Microsoft Q104011 (http://support.microsoft.com/default.aspx?scid=kb;EN-US;q104011)
*/

#include <string.h>
#include <windows.h>

/* #define EnvRegType REG_SZ */
#define EnvRegType REG_EXPAND_SZ

#define SYSENV  "SYSTEM\\CurrentControlSet\\Control\\Session Manager\\Environment"
#define USERENV "Environment"

#define PATH  "Path"

void Usage()
 {
  MessageBox(NULL,
             "Set persistent environment variable for NT and make it immediately available.\r\n"
	     "\r\n"
	     "Usage: SetEnvNT [/s] [/a|/r] Environment_var [Value]\r\n"
	     "  /s: System environment variable\r\n"
	     "  /a: Add Value. Only applicable if Environment_var is PATH.\r\n"
	     "  /r: Remove Value. Only applicable if Environment_var is PATH.\r\n"
	     " If Environment_var is PATH then Value is required.",
	     "SetEnvNT",
	     MB_OK|MB_ICONINFORMATION);
  exit(1);
 }

void *RegGetKeyValue(char *sKeyName,char *sValueName,HKEY WhichKey,long *RegType)
 {
  HKEY hKey;
  long lRetVal,cch,_RegType;
  void *Value=NULL;

  lRetVal=RegOpenKeyEx(WhichKey,sKeyName,0,KEY_READ,&hKey);
  if (lRetVal==0) {
   lRetVal=RegQueryValueEx(hKey,sValueName,NULL,&_RegType,NULL,&cch);
   if (lRetVal==0) {
    Value=malloc(cch);
    lRetVal=RegQueryValueEx(hKey,sValueName,NULL,&_RegType,Value,&cch);
    if (lRetVal) {
     free(Value);
     Value=NULL;
    }
   }
   RegCloseKey(hKey);
  }
  if ((Value!=NULL) && (RegType!=NULL))
   *RegType=_RegType;
  return(Value);
 }

void RegSetKeyValue(char *sKeyName,char *sValueName,void *ValueSetting,long RegType,HKEY WhichKey)
 {
  HKEY hKey;
  long lRetVal,cch;

  lRetVal=RegOpenKeyEx(WhichKey,sKeyName,0,KEY_ALL_ACCESS,&hKey);
  if (lRetVal==0) {
   switch (RegType) {
   case REG_SZ:
   case REG_EXPAND_SZ:
    cch=strlen(ValueSetting);
    break;
   case REG_DWORD:
    cch=sizeof(DWORD);
    break;
   default:
    cch=-1;
    break;
   }
   if (cch>=0)
    lRetVal=RegSetValueEx(hKey,sValueName,0,RegType,ValueSetting,cch);
   RegCloseKey(hKey);
  }
 }

void RegDelKeyValue(char *sKeyName,char *sValueName,HKEY WhichKey)
 {
  HKEY hKey;
  long lRetVal;

  lRetVal=RegOpenKeyEx(WhichKey,sKeyName,0,KEY_ALL_ACCESS,&hKey);
  if (lRetVal==0) {
   lRetVal=RegDeleteValue(hKey,sValueName);
   RegCloseKey(hKey);
  }
 }

void setarg(lpCmdLine,argc,argv)
 int *argc;
 char *lpCmdLine,***argv;
 {
  int i;
  char c,*pointer,*pointer1,_pgmname[MAX_PATH];

  pointer=lpCmdLine;
  for (*argc=1,i=0;*pointer;pointer++) {
   if (*pointer=='"') i=1-i;
   if ((i==0) && (isspace(*pointer))) {
    (*argc)++;
    while (isspace(*pointer)) pointer++;
    if (*pointer==0) break;
    pointer--;
   }
  }
  *argv=malloc(((*argc)+1)*sizeof(**argv));
  (*argv)[0]=NULL;
  if ((GetModuleFileName(NULL,_pgmname,sizeof(_pgmname))>0) &&
      ((pointer=strdup(_pgmname))!=NULL)) (*argv)[0]=pointer;
  if (*lpCmdLine) {
   pointer=pointer1=lpCmdLine;
   for (*argc=1,i=0;;pointer++) {
    if (*pointer=='"') i=1-i;
    if (((i==0) || (*pointer==0)) && ((isspace(*pointer)) || (*pointer==0))) {
     c=*pointer;
     *pointer=0;
     (*argv)[*argc]=strdup(pointer1);
     if (((*argv)[*argc][0]=='"') && ((*argv)[*argc][strlen((*argv)[*argc])-1]=='"')) {
      strcpy((*argv)[*argc],(*argv)[*argc]+1);
      (*argv)[*argc][strlen((*argv)[*argc])-1]=0;
     }
     (*argc)++;
     *pointer=c;
     while (isspace(*pointer)) pointer++;
     pointer1=pointer--;
     if (*pointer1==0) break;
    }
    if (*pointer==0) break;
   }
  }
 }

int InPath(char *opath, char *oval)
{
  int     ret = 0;
  char    *val, *path, *p;

  path = strupr(strdup(opath));
  val = strupr(strdup(oval));
  if (strstr(path, val)) {
    p = strtok(path, ";");

    if (!p && !strcmp(opath, val)) {
      ret = 1;
    }
    else {
      while (p != NULL) {
	if (strcmp(val, p)) {
	  ret = 1;
	  break;
	}
	p = strtok(NULL, ";");
      }
    }
  }
  free(val);
  free(path);

  return ret;
}

char *StripPath(char *opath, char *val)
{
  char *str, *path, *p;

  str  = (char *) calloc(1, strlen(opath));
  path = strdup(opath);

  p = strtok(path, ";");

  /* if it is a signle item on the path and
     eq to val then return empty str */

  if (p || strcmpi(opath, val)) {

    /* loop through each token and skip if it matches val */

    while (p != NULL) {
      if (strcmpi(val, p)) {
	strcat(str, p);
	strcat(str, ";");
      }
      p = strtok(NULL, ";");
    }

    /* get rid of trailing ";" */

    str[strlen(str)-1] = '\0';
  }
  free(path);

  return str;
}

int WINAPI WinMain(HINSTANCE hInstance,HINSTANCE hPrevInstance,LPSTR lpCmdLine,int nCmdShow)
 {
  long Result;
  int argc,i;
  char **argv,SystemVar=FALSE,*var=NULL,*value=NULL,*KeyName,add=FALSE,remove=FALSE,*v=NULL,*newstr=NULL;
  HKEY WhichKey;

  setarg(lpCmdLine,&argc,&argv);

  for (i=1;i<argc;i++)
   if (stricmp(argv[i],"/s")==0)
    SystemVar=TRUE;
   else if (stricmp(argv[i],"/a")==0)
    add=TRUE;
   else if (stricmp(argv[i],"/r")==0)
    remove=TRUE;
   else if (argv[i][0]=='/')
    Usage();
   else if (var==NULL)
    var=argv[i];
   else if (value==NULL)
    value=argv[i];
   else
    Usage();

  if ((var==NULL) || ((add) && (remove)))
   Usage();

  if (stricmp(var,PATH)) {
   if ((add) || (remove))
    Usage();
  }
  else {
   if (value==NULL)
    Usage();
  }

  if (SystemVar) {
   KeyName=SYSENV;
   WhichKey=HKEY_LOCAL_MACHINE;
  }
  else {
   KeyName=USERENV;
   WhichKey=HKEY_CURRENT_USER;
  }

  if ((add || remove)) {
   v=RegGetKeyValue(KeyName,var,WhichKey,NULL);

   if (add) 
    if (!InPath(v,value)) {
     newstr=(char *) calloc(1,strlen(v)+strlen(value)+2);
     strcpy(newstr,v);
     if (newstr[strlen(newstr)-1]!=';')
      strcat(newstr,";");
     strcat(newstr,value);
     value=newstr;
    }
    else
     value=v;

   if (remove) 
    if (InPath(v,value))
     value=StripPath(v,value);
    else
     value=v;
  }

  if (value!=NULL)
   RegSetKeyValue(KeyName,var,value,EnvRegType,WhichKey);
  else
   RegDelKeyValue(KeyName,var,WhichKey);

  SendMessageTimeout(HWND_BROADCAST,WM_SETTINGCHANGE,0,(long) USERENV,SMTO_ABORTIFHUNG,5000,&Result);

  if (v!=NULL)
   free(v);

  if (newstr!=NULL)
   free(newstr);

  return(0);
 }
